"""Tests for the `question_form_fields` module."""

from django.test import SimpleTestCase
from .. import question_form_fields


class TestIsChoiceField(SimpleTestCase):
    """Tests for the `is_choice_field` function."""

    def test_is_choice_field(self):
        """Test that the function returns True for the choice fields."""
        self.assertTrue(question_form_fields.is_choice_field("choice"))
        self.assertTrue(
            question_form_fields.is_choice_field("multiple_choice"),
        )

    def test_is_not_choice_field(self):
        """Test that the function returns False for non choice fields."""
        for field in question_form_fields.FieldTypeChoices:
            if field not in ("choice", "multiple_choice"):
                self.assertFalse(
                    question_form_fields.is_choice_field(field),
                    f"{field} is a choice field.",
                )



import typing as _t
from django import forms
from django.db import models


class FieldTypeChoices(models.TextChoices):
    """The type of field."""

    TEXT = "text", "Text"
    TEXTAREA = "textarea", "Textarea"
    EMAIL = "email", "Email"
    INTEGER = "integer", "Integer"
    DECIMAL = "decimal", "Decimal"
    FLOAT = "float", "Float"
    BOOLEAN = "boolean", "Boolean"
    DATE = "date", "Date"
    DATETIME = "datetime", "DateTime"
    TIME = "time", "Time"
    URL = "url", "URL"
    CHOICE = "choice", "Choice"
    MULTIPLE_CHOICE = "multiple_choice", "Multiple Choice"


field_type_map = {
    FieldTypeChoices.TEXT: forms.CharField,
    FieldTypeChoices.TEXTAREA: forms.CharField,
    FieldTypeChoices.EMAIL: forms.EmailField,
    FieldTypeChoices.INTEGER: forms.IntegerField,
    FieldTypeChoices.DECIMAL: forms.DecimalField,
    FieldTypeChoices.FLOAT: forms.FloatField,
    FieldTypeChoices.BOOLEAN: forms.BooleanField,
    FieldTypeChoices.DATE: forms.DateField,
    FieldTypeChoices.DATETIME: forms.DateTimeField,
    FieldTypeChoices.TIME: forms.TimeField,
    FieldTypeChoices.URL: forms.URLField,
    FieldTypeChoices.CHOICE: forms.ChoiceField,
    FieldTypeChoices.MULTIPLE_CHOICE: forms.MultipleChoiceField,
}


def is_choice_field(field: _t.Union[str, forms.Field]) -> bool:
    """Check if the field is a choice field.

    :param field: The field to check.
    :type field: str or forms.Field
    :returns: True if the field is a choice field, False otherwise.
    :rtype: bool
    """
    if isinstance(field, str):
        field = field_type_map[field]
    return issubclass(field.widget, forms.Select)



from datetime import timedelta
from django.test import TestCase
from django.utils import timezone
from model_bakery import baker
from .. import models as fc_models


class TestForms(TestCase):
    """Tests custom managers and querysets associated with the Form model."""

    def test_live_qs(self):
        """Test the live queryset only returns live forms."""

        # Inactive form
        baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.INACTIVE,
            start_dt=timezone.now() - timedelta(days=1),
            end_dt=None,
        )

        # Form that hasn't started yet
        baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.ACTIVE,
            start_dt=timezone.now() + timedelta(days=1),
            end_dt=None,
        )

        # Form that has ended
        baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.ACTIVE,
            start_dt=timezone.now() - timedelta(days=1, hours=1),
            end_dt=timezone.now() - timedelta(days=1),
        )

        # Live forms
        live_form_1 = baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.ACTIVE,
            start_dt=timezone.now() - timedelta(days=1),
            end_dt=timezone.now() + timedelta(days=1),
        )
        live_form_2 = baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.ACTIVE,
            start_dt=timezone.now() - timedelta(days=1),
            end_dt=None,
        )

        results = fc_models.Form.objects.filter().live()
        self.assertEqual(results.count(), 2)
        self.assertIn(live_form_1, results)
        self.assertIn(live_form_2, results)

    def test_live_manager_method(self):
        """Test the manager's `live` method."""
        baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.INACTIVE,
        )
        baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.ACTIVE,
            start_dt=timezone.now() - timedelta(days=1),
            end_dt=timezone.now() + timedelta(days=1),
        )

        results = set(
            fc_models.Form.objects.live().values_list("id", flat=True)
        )
        expected_results = set(
            fc_models.Form.objects.filter().live().values_list("id", flat=True)
        )
        self.assertEqual(results, expected_results)



import typing as _t
import sys
from django.db import models
from django.db.models import Q, QuerySet
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.utils import timezone
from django.utils.text import slugify
from django.urls import reverse
from .question_form_fields import FieldTypeChoices, is_choice_field
from .managers import FormManager

User = get_user_model()

# When running test suite, it moans about the url app name not existing in the
# namespace. This fixes it, but it's a bit of a hack.
TESTING = "test" in sys.argv[0]
url_prefix = "form_creator:"


class Form(models.Model):
    """The configuration for a form."""

    class StatusChoices(models.TextChoices):
        """The status of a form."""

        DRAFT = "draft", "Draft"
        ACTIVE = "active", "Active"
        INACTIVE = "inactive", "Inactive"

    owner = models.ForeignKey(User, on_delete=models.CASCADE)
    editors = models.ManyToManyField(User, related_name="editors", blank=True)
    title = models.CharField(max_length=150)
    slug = models.SlugField(max_length=150)
    description = models.TextField(blank=True)
    created_dt = models.DateTimeField(auto_now_add=True)
    start_dt = models.DateTimeField(
        default=timezone.now,
        verbose_name="Start on",
    )
    end_dt = models.DateTimeField(
        blank=True,
        null=True,
        help_text="Leave blank for no end date.",
        verbose_name="End on",
    )
    status = models.CharField(
        max_length=10,
        choices=StatusChoices.choices,
        default=StatusChoices.DRAFT,
    )

    objects = FormManager()

    class Meta:
        db_table = "fc_form"
        ordering = ["status", "-created_dt"]

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        """Override the save method to set the slug."""
        if not self.slug:
            self.slug = slugify(self.title)
        super().save(*args, **kwargs)

    def can_edit(self, user: User) -> bool:
        """Check if the user can edit the form."""
        if not user or not user.is_authenticated:
            return False
        return any(
            [
                user.is_staff,
                user.username == self.owner.username,
                user in self.editors.all(),
            ]
        )

    def can_delete(self, user: User) -> bool:
        """Check if the user can delete the form."""
        if not user or not user.is_authenticated:
            return False
        return any([user.is_staff, user.username == self.owner.username])

    def completed_by(self, user: User) -> _t.Optional["FormResponder"]:
        """Get the form responder for the user."""
        if not user or not user.is_authenticated:
            return None
        return self.responders.filter(user=user).first()

    def can_complete_form(self, user: User) -> bool:
        """Check if the user can complete the form."""
        if not user or not user.is_authenticated:
            return False

        if not self.is_live():
            return False

        if self.can_edit(user):
            return False

        return not bool(self.completed_by(user))

    def get_absolute_url(self) -> str:
        """Get the absolute URL for the form."""
        return reverse(f"{url_prefix}form_detail", args=[self.id, self.slug])

    def get_edit_url(self) -> str:
        """Get the URL for the form's edit view."""
        return reverse(f"{url_prefix}form_edit", args=[self.id, self.slug])

    def get_delete_url(self) -> str:
        """Get the URL for the form's delete view."""
        return reverse(f"{url_prefix}form_delete", args=[self.id, self.slug])

    def get_edit_questions_url(self) -> str:
        """Get the URL for the form's questions edit view."""
        return reverse(
            f"{url_prefix}form_questions_edit",
            args=[self.id, self.slug],
        )

    def get_respond_url(self) -> str:
        """Get the URL to start filling out the form."""
        return reverse(f"{url_prefix}form_response", args=[self.id, self.slug])

    @classmethod
    def get_editable_forms(cls, user: _t.Optional[User]) -> QuerySet["Form"]:
        """Get the forms that the user can edit."""
        if not user or not user.is_authenticated:
            return cls.objects.none()
        return cls.objects.filter(Q(owner=user) | Q(editors=user))

    def is_live(self) -> bool:
        """Indicate if the form is live."""
        if not self.status == self.StatusChoices.ACTIVE:
            return False

        if self.end_dt and self.end_dt < timezone.now():
            return False

        if self.start_dt > timezone.now():
            return False

        return True

    @property
    def num_responses(self) -> int:
        """Get the number of responses for the form."""
        return self.responders.count()


class FormQuestion(models.Model):
    """A collection of questions for a form."""

    form = models.ForeignKey(
        Form,
        on_delete=models.CASCADE,
        related_name="questions",
    )
    field_type = models.CharField(
        max_length=32,
        choices=FieldTypeChoices.choices,
        default=FieldTypeChoices.TEXT,
    )
    question = models.CharField(max_length=150)
    description = models.TextField(blank=True, null=True)
    required = models.BooleanField(default=False)
    seq_no = models.IntegerField(default=0)
    choices = models.TextField(
        blank=True,
        null=True,
        help_text="Separate choices with a pipe (|).",
    )
    related_question = models.ForeignKey(
        "self",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="The question to use for the related question.",
    )

    class Meta:
        db_table = "fc_form_question"
        ordering = ["form", "seq_no"]
        unique_together = ["form", "question"]

    def __str__(self):
        return f"{self.form.title} - {self.question}"

    @property
    def choice_list(self) -> _t.List[str]:
        """Get the list of choices for the question."""
        return (self.choices or "").split("|")

    def clean(self, *args, **kwargs) -> None:
        """Ensure that fields that require choices are not blank and
        vice-versa.
        """
        super().clean(*args, **kwargs)
        _is_choice_field = is_choice_field(self.field_type)
        choices = (self.choices or "").strip()
        if _is_choice_field and not choices:
            raise ValidationError("This question field type requires choices.")
        if not _is_choice_field and choices:
            raise ValidationError(
                "This question field type does not support choices."
            )


class FormResponder(models.Model):
    """Represents a person responding to a form."""

    form = models.ForeignKey(
        Form,
        on_delete=models.CASCADE,
        related_name="responders",
    )
    created_dt = models.DateTimeField(auto_now_add=True)
    user = models.ForeignKey(User, on_delete=models.CASCADE)

    class Meta:
        db_table = "fc_form_responder"
        ordering = ["-created_dt"]
        unique_together = ["form", "user"]

    def __str__(self):
        return f"{self.form.title} - {self.created_dt}"


class FormResponse(models.Model):
    """A response to a form."""

    form_responder = models.ForeignKey(
        FormResponder,
        on_delete=models.CASCADE,
        related_name="responses",
    )
    question = models.ForeignKey(
        FormQuestion,
        on_delete=models.CASCADE,
    )
    answer = models.TextField(blank=True, null=True)

    class Meta:
        db_table = "fc_form_response"
        ordering = ["form_responder", "question"]

    def __str__(self):
        return f"{self.form_responder.form.title} - {self.question.question}"



from django.test import TestCase
from django import forms
from django.forms import ValidationError
from django.contrib.auth import get_user_model
from model_bakery import baker
from .. import forms as fc_forms, models as fc_models
from ..question_form_fields import FieldTypeChoices

User = get_user_model()


class TestNewForm(TestCase):
    def test_form_loads(self):
        """Test that the form actually loads."""
        baker.make(User, _quantity=2)
        editors = User.objects.all()
        owner = baker.make(User)
        form = fc_forms.NewForm(
            user=owner,
            editor_choices=editors,
        )

        self.assertEqual(form.fields["owner"].initial, owner)
        self.assertEqual(
            set(form.fields["editors"].queryset.values_list("id")),
            set(editors.values_list("id")),
        )

    def test_form_no_editors(self):
        """Test the form load with no editors."""
        form = fc_forms.NewForm(
            user=baker.make(User),
        )

        self.assertEqual(
            set(form.fields["editors"].queryset.values_list("id")),
            set(),
        )


class TestDeleteForm(TestCase):
    def test_form_loads(self):
        """Test that the form actually loads."""
        form = baker.make(fc_models.Form)
        fc_forms.DeleteForm(instance=form, user=form.owner)

    def test_clean_user(self):
        """Test that the `clean` method accepts a valid user."""

        form = baker.make(fc_models.Form)
        fc_forms.DeleteForm(instance=form, user=form.owner)
        form.clean_fields()

    def test_clean_user_invalid(self):
        """Test that the `clean` method rejects an invalid user."""

        with self.assertRaises(ValidationError):
            form = fc_forms.DeleteForm(
                instance=baker.make(fc_models.Form),
                user=baker.make(User),
            )
            form.clean()


class TestFormQuestionForm(TestCase):
    """Test the FormQuestionForm."""

    def test_init(self):
        """Test the initialiser ensuring that the form loads correctly."""
        form = fc_forms.FormQuestionForm(1)
        self.assertEqual(form.form_id, 1)

    def test_save_existing_form(self):
        """Test that an existing form saves correctly."""
        form_1_obj = baker.make(fc_models.Form, title="q1")
        form_2_obj = baker.make(fc_models.Form, title="q2")
        form_question_obj = baker.make(fc_models.FormQuestion, form=form_1_obj)
        form_question = fc_forms.FormQuestionForm(
            form_id=form_2_obj.id,
            data=form_question_obj.__dict__,
        )
        form_question.is_valid()
        saved_obj = form_question.save()
        self.assertEqual(saved_obj.form, form_2_obj)

    def test_save_new_form(self):
        """Test that a new form saves correctly."""
        form_obj = baker.make(fc_models.Form, title="q1")
        form = fc_forms.FormQuestionForm(
            form_id=form_obj.id,
            initial={
                "question": "q1",
            },
        )
        form.is_valid()
        form.save()
        self.assertEqual(fc_models.FormQuestion.objects.count(), 1)


class TestCaptureResponseForm(TestCase):
    """Test the CaptureResponseForm."""

    @classmethod
    def setUpTestData(cls):
        """Set up the test data."""
        cls.form = baker.make(fc_models.Form)
        cls.text_q = baker.make(fc_models.FormQuestion, form=cls.form)
        cls.choice_q = baker.make(
            fc_models.FormQuestion,
            form=cls.form,
            field_type=FieldTypeChoices.CHOICE,
            choices="a|b|c",
        )

    def test_init(self):
        """Test the initialiser ensuring that the form loads correctly."""
        form = fc_forms.CaptureResponseForm(self.form)
        self.assertEqual(form.form, self.form)

    def test_loads_fields(self):
        """Test that the form loads the correct fields."""
        form = fc_forms.CaptureResponseForm(self.form)
        self.assertEqual(len(form.fields), 2)

        text_q_field = form.fields["question_%s" % self.text_q.id]
        choice_q_field = form.fields["question_%s" % self.choice_q.id]

        self.assertEqual(text_q_field.label, self.text_q.question)
        self.assertEqual(choice_q_field.label, self.choice_q.question)

        self.assertIsInstance(text_q_field.widget, forms.TextInput)
        self.assertIsInstance(choice_q_field.widget, forms.Select)

    def test_load_choices(self):
        """Test that form loads the choices correctly for the choice field."""
        form = fc_forms.CaptureResponseForm(self.form)
        choice_q_field = form.fields["question_%s" % self.choice_q.id]
        self.assertEqual(
            choice_q_field.widget.choices,
            [("a", "a"), ("b", "b"), ("c", "c")],
        )

    def test_save(self):
        """Test that the form saves correctly."""
        form = fc_forms.CaptureResponseForm(
            self.form,
            data={
                "question_%s" % self.text_q.id: "q1",
                "question_%s" % self.choice_q.id: "a",
            },
        )
        form.is_valid()

        form_response = form.save(baker.make(User))

        self.assertEqual(fc_models.FormResponder.objects.count(), 1)
        self.assertEqual(
            fc_models.FormResponder.objects.first(),
            form_response,
        )
        self.assertEqual(form_response.responses.count(), 2)
        self.assertEqual(
            set(
                form_response.responses.all().values_list(
                    "question_id", flat=True
                )
            ),
            {self.text_q.id, self.choice_q.id},
        )
        self.assertEqual(
            form_response.responses.get(question=self.text_q).answer,
            "q1",
        )
        self.assertEqual(
            form_response.responses.get(question=self.choice_q).answer,
            "a",
        )



from django.db import models
from django.utils import timezone


class FormsQueryset(models.QuerySet):
    """QuerySet for the Form model."""

    def live(self):
        """Return only live forms."""
        now = timezone.now()
        # Ideally we would import the active status, but that would cause a
        # circular import.
        return (
            self.filter(status="active")
            .exclude(start_dt__gt=now)
            .exclude(end_dt__lt=now)
        )


class FormManager(models.Manager):
    """Manager for the Form model."""

    def live(self):
        """Return only live forms."""
        return self.get_queryset().live()

    def get_queryset(self):
        """Return a queryset for the Form model."""
        return FormsQueryset(self.model, using=self._db)



"""Tests for the `form_creator_tags` module."""

from django.test import TestCase
from django.contrib.auth import get_user_model
from model_bakery import baker
from .. import models as fc_models
from ..templatetags import form_creator_tags


User = get_user_model()


class TestCanEditForm(TestCase):
    """Tests for the `can_edit_form` filter."""

    def setUp(self):
        self.user = baker.make(User)
        self.form = baker.make(fc_models.Form)

    def test_can_edit_form_true(self):
        """Test that the filter returns True when the user is the owner."""
        self.form.owner = self.user
        self.form.save()
        self.assertTrue(form_creator_tags.can_edit_form(self.form, self.user))

    def test_can_edit_form_false(self):
        """Test that the filter returns False when the user is not the
        owner.
        """
        self.assertFalse(form_creator_tags.can_edit_form(self.form, self.user))


class TestCanDeleteForm(TestCase):
    """Tests for the `can_delete_form` filter."""

    def setUp(self):
        self.user = baker.make(User)
        self.form = baker.make(fc_models.Form)

    def test_can_delete_form_true(self):
        """Test that the filter returns True when the user is the owner."""
        self.form.owner = self.user
        self.form.save()
        self.assertTrue(
            form_creator_tags.can_delete_form(self.form, self.user)
        )

    def test_can_delete_form_false(self):
        """Test that the filter returns False when the user is not the
        owner.
        """
        self.assertFalse(
            form_creator_tags.can_delete_form(self.form, self.user)
        )


class TestCanCompleteForm(TestCase):
    """Tests for the `can_complete_form` filter."""

    def setUp(self):
        self.user = baker.make(User)
        self.form = baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.ACTIVE,
        )

    def test_can_complete_form_true(self):
        """Test that the filter returns True when the user is the owner."""
        self.assertTrue(
            form_creator_tags.can_complete_form(self.form, self.user)
        )

    def test_can_complete_form_false(self):
        """Test that the filter returns False when the user is not the
        owner.
        """
        self.form.owner = self.user
        self.form.save()
        self.assertFalse(
            form_creator_tags.can_complete_form(self.form, self.user)
        )


class TestFormCompletedOn(TestCase):
    """Tests for the `form_completed_on` filter."""

    def setUp(self):
        self.user = baker.make(User)
        self.form = baker.make(fc_models.Form)

    def test_form_not_completed(self):
        """Test that the filter returns an empty string when the form is not
        completed.
        """
        self.assertEqual(
            form_creator_tags.form_completed_on(self.form, self.user),
            "",
        )

    def test_form_completed(self):
        """Test that the filter returns a string representation of a date
        when the form is completed.
        """
        responder = baker.make(
            fc_models.FormResponder, form=self.form, user=self.user
        )
        self.assertEqual(
            form_creator_tags.form_completed_on(self.form, self.user),
            responder.created_dt.strftime("%Y-%m-%d"),
        )



# Generated by Django 4.0.4 on 2022-06-16 18:47

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Form",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=150)),
                ("slug", models.SlugField(max_length=150)),
                ("description", models.TextField(blank=True)),
                ("created_dt", models.DateTimeField(auto_now_add=True)),
                (
                    "start_dt",
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        verbose_name="Start on",
                    ),
                ),
                (
                    "end_dt",
                    models.DateTimeField(
                        blank=True,
                        help_text="Leave blank for no end date.",
                        null=True,
                        verbose_name="End on",
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("draft", "Draft"),
                            ("active", "Active"),
                            ("inactive", "Inactive"),
                        ],
                        default="draft",
                        max_length=10,
                    ),
                ),
                (
                    "editors",
                    models.ManyToManyField(
                        blank=True,
                        related_name="editors",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "fc_form",
                "ordering": ["status", "-created_dt"],
            },
        ),
        migrations.CreateModel(
            name="FormQuestion",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "field_type",
                    models.CharField(
                        choices=[
                            ("text", "Text"),
                            ("textarea", "Textarea"),
                            ("email", "Email"),
                            ("integer", "Integer"),
                            ("decimal", "Decimal"),
                            ("float", "Float"),
                            ("boolean", "Boolean"),
                            ("date", "Date"),
                            ("datetime", "DateTime"),
                            ("time", "Time"),
                            ("url", "URL"),
                            ("choice", "Choice"),
                            ("multiple_choice", "Multiple Choice"),
                        ],
                        default="text",
                        max_length=32,
                    ),
                ),
                ("question", models.CharField(max_length=150)),
                ("description", models.TextField(blank=True, null=True)),
                ("required", models.BooleanField(default=False)),
                ("seq_no", models.IntegerField(default=0)),
                (
                    "choices",
                    models.TextField(
                        blank=True,
                        help_text="Separate choices with a pipe (|).",
                        null=True,
                    ),
                ),
                (
                    "form",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="questions",
                        to="form_creator.form",
                    ),
                ),
                (
                    "related_question",
                    models.ForeignKey(
                        blank=True,
                        help_text="The question to use for the related question.",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="form_creator.formquestion",
                    ),
                ),
            ],
            options={
                "db_table": "fc_form_question",
                "ordering": ["form", "seq_no"],
            },
        ),
        migrations.CreateModel(
            name="FormResponder",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_dt", models.DateTimeField(auto_now_add=True)),
                (
                    "form",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="responders",
                        to="form_creator.form",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "fc_form_responder",
                "ordering": ["-created_dt"],
            },
        ),
        migrations.CreateModel(
            name="FormResponse",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("answer", models.TextField(blank=True, null=True)),
                (
                    "form_responder",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="form_creator.formresponder",
                    ),
                ),
                (
                    "question",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="form_creator.formquestion",
                    ),
                ),
            ],
            options={
                "db_table": "fc_form_response",
                "ordering": ["form_responder", "question"],
            },
        ),
    ]



from datetime import timedelta
from django.test import TestCase
from django.utils import timezone
from django.core.exceptions import ValidationError
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from model_bakery import baker
from .. import models as fc_models
from ..question_form_fields import FieldTypeChoices


User = get_user_model()


class TestForm(TestCase):
    """Test the Form model."""

    def test_str(self):
        """Test that the `__str__` method returns a string instance."""
        self.assertIsInstance(str(baker.make(fc_models.Form)), str)

    def test_can_edit_owner(self):
        """Test that the owner can edit the form."""
        form = baker.make(fc_models.Form, owner=baker.make(User))
        self.assertTrue(form.can_edit(form.owner))

    def test_can_edit_editors(self):
        """Test that editors can edit the form."""
        form = baker.make(
            fc_models.Form,
            editors=baker.make(User, _quantity=2),
        )
        self.assertTrue(form.can_edit(form.editors.first()))

    def test_can_edit_other(self):
        """Test that other users cannot edit the form."""
        form = baker.make(fc_models.Form)
        self.assertFalse(form.can_edit(baker.make(User)))

    def test_can_edit_anon_user(self):
        """Test that the `can_edit` method returns `False` for an anonymous
        user.
        """
        form = baker.make(fc_models.Form)
        self.assertFalse(form.can_edit(AnonymousUser()))

    def test_completed_by(self):
        """Test that the `completed_by` method returns the correct instance of
        `FormResponder`.
        """
        form = baker.make(fc_models.Form)
        user = baker.make(User)
        form_responder = baker.make(
            fc_models.FormResponder,
            form=form,
            user=user,
        )
        baker.make(fc_models.FormResponder, form=form)
        self.assertEqual(form.completed_by(user), form_responder)

    def test_completed_by_not_completed(self):
        """Test that the `completed_by` method returns `None` when the user
        has not completed the form.
        """
        user = baker.make(User)
        form = baker.make(fc_models.Form)
        baker.make(fc_models.FormResponder, form=form)
        self.assertIsNone(form.completed_by(user))

    def test_completed_by_none_user(self):
        """Test that the `completed_by` method returns `None` when the user
        is `None`.
        """
        form = baker.make(fc_models.Form)
        self.assertIsNone(form.completed_by(None))

    def test_can_complete_form(self):
        """Test that the `can_complete_form` method returns the correct
        boolean.
        """
        form = baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.ACTIVE,
        )
        user = baker.make(User)
        self.assertTrue(form.can_complete_form(user))

    def test_can_complete_form_not_live(self):
        """Test that the `can_complete_form` method returns the correct
        boolean when the form is not live.
        """
        form = baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.DRAFT,
        )
        user = baker.make(User)
        self.assertFalse(form.can_complete_form(user))

    def test_can_complete_form_editor(self):
        """Test that the `can_complete_form` method returns the correct
        boolean for an editor.
        """
        user = baker.make(User)
        form = baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.ACTIVE,
        )
        form.editors.add(user)
        self.assertFalse(form.can_complete_form(user))

    def test_can_complete_form_completed(self):
        """Test that the `can_complete_form` method returns the correct
        boolean when the form is completed by the user.
        """
        form = baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.ACTIVE,
        )
        user = baker.make(User)
        baker.make(fc_models.FormResponder, form=form, user=user)
        self.assertFalse(form.can_complete_form(user))

    def test_can_complete_form_anon_user(self):
        """Test that the `can_complete_form` method returns `False` for an
        anonymous user.
        """
        form = baker.make(fc_models.Form)
        self.assertFalse(form.can_complete_form(AnonymousUser()))

    def test_get_absolute_url(self):
        """Test that the `get_absolute_url` method returns a string
        instance.
        """
        self.assertIsInstance(
            baker.make(fc_models.Form).get_absolute_url(),
            str,
        )

    def test_get_edit_url(self):
        """Test that the `get_edit_url` method returns a string instance."""
        self.assertIsInstance(
            baker.make(fc_models.Form).get_edit_url(),
            str,
        )

    def test_get_delete_url(self):
        """Test that the `get_delete_url` method returns a string instance."""
        self.assertIsInstance(
            baker.make(fc_models.Form).get_delete_url(),
            str,
        )

    def test_get_edit_questions_url(self):
        """Test that the `get_edit_questions_url` method returns a string
        instance.
        """
        self.assertIsInstance(
            baker.make(fc_models.Form).get_edit_questions_url(),
            str,
        )

    def test_get_respond_url(self):
        """Test that the `get_respond_url` method returns a string instance."""
        self.assertIsInstance(
            baker.make(fc_models.Form).get_respond_url(),
            str,
        )

    def test_save_slug(self):
        """Test that the `save` method sets the slug."""
        form = baker.make(fc_models.Form, title="Test Form", slug="")
        self.assertEqual(form.slug, "test-form")

    def test_is_live_is_live(self):
        """Test that the `is_live` method returns `True` when the form is
        live.
        """

        form = baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.ACTIVE,
            start_dt=timezone.now() - timedelta(days=1),
            end_dt=timezone.now() + timedelta(days=1),
        )
        self.assertTrue(form.is_live())

    def test_is_live_start_dt(self):
        """Test that the `is_live` returns the correct boolean when tweaking
        the start date.
        """

        form = baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.ACTIVE,
            start_dt=timezone.now() + timedelta(days=1),
            end_dt=timezone.now() + timedelta(days=1),
        )
        self.assertFalse(form.is_live())

    def test_is_live_end_dt(self):
        """Test that the `is_live` returns the correct boolean when tweaking
        the end date.
        """

        # End date in past
        form = baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.ACTIVE,
            start_dt=timezone.now() - timedelta(days=1),
            end_dt=timezone.now() - timedelta(days=1),
        )
        self.assertFalse(form.is_live())

        # No end date
        form.end_dt = None
        self.assertTrue(form.is_live())

    def test_is_live_status(self):
        """Test that the `is_live` returns the correct boolean when tweaking
        the status.
        """

        form = baker.make(
            fc_models.Form,
            status=fc_models.Form.StatusChoices.DRAFT,
            start_dt=timezone.now() - timedelta(days=1),
            end_dt=timezone.now() + timedelta(days=1),
        )
        self.assertFalse(form.is_live())

    def test_get_editable_forms_anon_user(self):
        """Test that the `get_editable_forms` method returns an empty queryset
        for an anonymous user.
        """
        self.assertQuerysetEqual(fc_models.Form.get_editable_forms(None), [])
        self.assertQuerysetEqual(
            fc_models.Form.get_editable_forms(AnonymousUser()),
            [],
        )

    def test_get_editable_forms(self):
        """Test that the `get_editable_forms` method returns a queryset of
        forms that the user can edit.
        """
        user = baker.make(User)
        owned_form = baker.make(fc_models.Form, owner=user)
        editable_form = baker.make(fc_models.Form, editors=[user])
        baker.make(fc_models.Form)

        results = fc_models.Form.get_editable_forms(user)
        self.assertEqual(results.count(), 2)
        self.assertIn(owned_form, results)
        self.assertIn(editable_form, results)

    def test_can_delete_anon_user(self):
        """Test that the `can_delete` method returns `False` for an
        anonymous user.
        """
        self.assertFalse(baker.make(fc_models.Form).can_delete(None))
        self.assertFalse(
            baker.make(fc_models.Form).can_delete(AnonymousUser())
        )

    def test_can_delete(self):
        """Test that the `can_delete` method returns `True` for the
        owner.
        """
        form = baker.make(fc_models.Form, owner=baker.make(User))
        self.assertTrue(form.can_delete(form.owner))

    def test_num_responses(self):
        """Test that the `num_responses` method returns the correct
        number of responses.
        """
        form = baker.make(fc_models.Form)
        baker.make(fc_models.FormResponder, form=form)
        baker.make(fc_models.FormResponder, form=form)
        self.assertEqual(form.num_responses, 2)


class TestFormQuestion(TestCase):
    """Test the FormQuestion model."""

    def test_str(self):
        """Test that the `__str__` method returns a string instance."""
        self.assertIsInstance(str(baker.make(fc_models.FormQuestion)), str)

    def test_choice_list(self):
        """Test that the `choice_list` method returns a list of choices."""
        form = baker.make(fc_models.FormQuestion, choices="a|b")
        self.assertEqual(form.choice_list, ["a", "b"])

    def test_clean_no_choice(self):
        """Test that the `clean` method raises an error when there are no
        choices but a field type requiring choices is selected.
        """
        form = baker.make(
            fc_models.FormQuestion,
            field_type=FieldTypeChoices.CHOICE,
            choices="",
        )
        with self.assertRaises(ValidationError):
            form.clean()

    def test_clean_non_choice_field(self):
        """Test that the `clean` method raises an error when there are choices
        but a field type not requiring choices is selected.
        """
        form = baker.make(
            fc_models.FormQuestion,
            field_type=FieldTypeChoices.TEXT,
            choices="a|b",
        )
        with self.assertRaises(ValidationError):
            form.clean()


class TestFormResponder(TestCase):
    """Test the FormResponder model."""

    def test_str(self):
        """Test that the `__str__` method returns a string instance."""
        self.assertIsInstance(str(baker.make(fc_models.FormResponder)), str)


class TestFormResponse(TestCase):
    """Test the FormResponse model."""

    def test_str(self):
        """Test that the `__str__` method returns a string instance."""
        self.assertIsInstance(str(baker.make(fc_models.FormResponse)), str)



import re
from django.shortcuts import get_object_or_404, render, redirect
from django.contrib import messages
from django.http import HttpRequest, HttpResponse
from django.urls import reverse_lazy
from django.forms import modelformset_factory
from django.views import View
from django.core.exceptions import PermissionDenied
from django.db.models import QuerySet
from django.utils.decorators import method_decorator
from django.views.generic.list import ListView
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView, DeleteView
from django.contrib.auth.decorators import login_required
from . import models as fc_models, forms as fc_forms, exporters as fc_exporters
from .decorators import with_form, redirect_if_form_completed


class FormBaseView(View):
    model = fc_models.Form
    form_class = fc_forms.NewForm
    editor_choices = None
    success_url = reverse_lazy("form_creator:form_list")

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        if self.editor_choices:
            kwargs["editor_choices"] = self.editor_choices
        return kwargs


class FormSingleItemMixin:
    def get_object(self, *args, **kwargs):
        pattern = re.compile("/forms/(\\d{1,})-(.*?)/")
        pk, slug = pattern.search(self.request.path).groups()
        return get_object_or_404(self.model, pk=pk, slug=slug)

    def get_context_data(self, **kwargs):
        """Adds permissions to the context."""
        context = super().get_context_data(**kwargs)
        obj = self.get_object()
        context["can_edit"] = obj.can_edit(self.request.user)
        context["can_delete"] = obj.can_delete(self.request.user)
        context["completed_form"] = obj.completed_by(self.request.user)
        context["can_complete_form"] = obj.can_complete_form(self.request.user)
        return context


class FormListView(FormBaseView, ListView):
    """List all forms. A owner and editor can see forms of all statuses.
    All other users can only see live forms.
    """

    def get_queryset(self) -> QuerySet[fc_models.Form]:
        """If the user is a staff member, return all forms. Otherwise, return
        only live forms and those which the user can edit/owns.
        """
        if self.request.user.is_staff:
            return self.model.objects.all()

        # Forms that are editable by the user
        qs = self.model.get_editable_forms(self.request.user)

        # Forms that are live
        qs |= self.model.objects.live()

        return qs


class FormCreateView(FormBaseView, CreateView):
    """View to create a new form."""

    template_name = "form_creator/form_create.html"


class FormDetailView(FormBaseView, FormSingleItemMixin, DetailView):
    """View to display a form."""

    template_name = "form_creator/form_detail.html"


class FormUpdateView(FormBaseView, FormSingleItemMixin, UpdateView):
    """View to edit a form."""

    template_name = "form_creator/form_edit.html"


class FormDeleteView(FormBaseView, FormSingleItemMixin, DeleteView):
    """View to delete a form."""

    form_class = fc_forms.DeleteForm
    template_name = "form_creator/form_delete.html"

    def get(self, request, *args, **kwargs):
        res = super().get(request, *args, **kwargs)
        if not self.object.can_delete(request.user):
            raise PermissionDenied(
                "You must be the owner to delete this form."
            )
        return res

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.pop("editor_choices", None)
        kwargs["instance"] = self.get_object()
        return kwargs


class FormQuestionsEditView(View):
    """View to manage questions for a form. Within this view the user would be
    able to add/edit/delete questions.
    """

    template_name = "form_creator/form_questions_edit.html"

    @method_decorator(with_form(can_edit=True), name="dispatch")
    def get(self, request: HttpRequest, form: fc_models.Form) -> HttpResponse:

        extra = 0 if form.questions.count() else 1
        FormQuestionFS = modelformset_factory(
            fc_models.FormQuestion,
            form=fc_forms.FormQuestionForm,
            extra=extra,
            can_delete=True,
        )

        return render(
            request,
            self.template_name,
            {
                "object": form,
                "formset": FormQuestionFS(
                    initial=form.questions.all().values(),
                    form_kwargs={"form_id": form.id},
                ),
            },
        )

    @method_decorator(with_form(can_edit=True), name="dispatch")
    def post(self, request: HttpRequest, form: fc_models.Form) -> HttpResponse:

        extra = 0 if form.questions.count() else 1
        FormQuestionFS = modelformset_factory(
            fc_models.FormQuestion,
            form=fc_forms.FormQuestionForm,
            extra=extra,
            can_delete=True,
        )

        formset = FormQuestionFS(
            request.POST, form_kwargs={"form_id": form.id}
        )
        if not formset.has_changed():
            return redirect(form.get_absolute_url())

        if formset.is_valid():
            formset.save()
            messages.success(request, "Questions edited.")
            return redirect(form.get_absolute_url())
        else:
            messages.error(request, "Please correct the errors below.")
            return render(
                request,
                self.template_name,
                {
                    "object": form,
                    "formset": formset,
                },
            )


class FormResponseView(View):
    """View for users to respond to questions in a form."""

    template_name = "form_creator/form_response.html"
    success_url = None

    @method_decorator(login_required, name="dispatch")
    @method_decorator(with_form(), name="dispatch")
    @method_decorator(redirect_if_form_completed(), name="dispatch")
    def get(self, request: HttpRequest, form: fc_models.Form) -> HttpResponse:
        return render(
            request,
            self.template_name,
            {"object": form, "form": fc_forms.CaptureResponseForm(form)},
        )

    @method_decorator(login_required, name="dispatch")
    @method_decorator(with_form(), name="dispatch")
    @method_decorator(redirect_if_form_completed(), name="dispatch")
    def post(self, request: HttpRequest, form: fc_models.Form) -> HttpResponse:
        response_form = fc_forms.CaptureResponseForm(form, request.POST)
        if response_form.is_valid():
            response_form.save(request.user)
            messages.success(request, "Response saved.")
            return redirect(self.success_url or form.get_absolute_url())
        else:
            messages.error(request, "Please correct the errors below.")
            return render(
                request,
                self.template_name,
                {"object": form, "form": response_form},
            )


@with_form(can_edit=True)
def download_questions(
    request: HttpRequest, form: fc_models.Form
) -> HttpResponse:
    """View to download a form's questions as a CSV file."""
    response = HttpResponse(content_type="text/csv")
    response["Content-Disposition"] = "attachment; filename=questions.csv"
    fc_exporters.export_questions(
        fc_models.FormQuestion.objects.filter(form=form),
        response,
    )
    return response


@with_form(can_edit=True)
def download_responses(
    request: HttpRequest, form: fc_models.Form
) -> HttpResponse:
    """View to download a form's responses as a CSV file."""
    response = HttpResponse(content_type="text/csv")
    response["Content-Disposition"] = "attachment; filename=responses.csv"
    fc_exporters.export_responses(
        fc_models.FormResponse.objects.filter(form_responder__form=form),
        response,
    )
    return response



from django.contrib import admin
from django.db import models
from django.db.models import QuerySet
from django.http import HttpRequest, HttpResponse
from django.forms import Textarea
from django.urls import resolve
from . import models as fc_models, exporters as fc_exporters


class TextAreaFormFieldOverride:
    formfield_overrides = {
        models.TextField: {"widget": Textarea(attrs={"rows": 1, "cols": 70})},
    }


class FormQuestionInline(TextAreaFormFieldOverride, admin.StackedInline):
    model = fc_models.FormQuestion
    extra = 0
    classes = ["collapse"]

    def formfield_for_foreignkey(self, db_field, request=None, **kwargs):
        if db_field.name == "related_question":
            obj_id = resolve(request.path).kwargs.get("object_id")
            if obj_id:
                kwargs["queryset"] = fc_models.FormQuestion.objects.filter(
                    form_id=obj_id
                )
            else:
                kwargs["queryset"] = fc_models.FormQuestion.objects.none()

        return super().formfield_for_foreignkey(db_field, request, **kwargs)


class FormResponderInline(TextAreaFormFieldOverride, admin.TabularInline):
    model = fc_models.FormResponder
    extra = 0
    raw_id_fields = ("user",)
    fields = (
        "id",
        "user",
        "created_dt",
    )
    readonly_fields = (
        "user",
        "created_dt",
    )
    show_change_link = True


class FormResponseInline(TextAreaFormFieldOverride, admin.TabularInline):
    model = fc_models.FormResponse
    extra = 0
    fields = ("question", "answer")
    raw_id_fields = ("question",)


@admin.register(fc_models.Form)
class FormAdmin(admin.ModelAdmin):
    list_display = ("title", "owner", "start_dt", "end_dt", "num_responses")
    list_filter = ("owner",)
    search_fields = ("title", "description")
    date_hierarchy = "start_dt"
    ordering = ("-start_dt",)
    raw_id_fields = ("owner",)
    prepopulated_fields = {"slug": ("title",)}
    readonly_fields = ("num_responses",)
    inlines = (FormQuestionInline, FormResponderInline)
    actions = ["export_questions", "export_responses"]
    fieldsets = (
        (
            None,
            {
                "fields": (
                    "title",
                    "slug",
                    "num_responses",
                    "description",
                    "owner",
                    "status",
                    "start_dt",
                    "end_dt",
                )
            },
        ),
        ("Editors", {"fields": ("editors",)}),
    )

    @admin.action(description="Export questions")
    def export_questions(
        self,
        request: HttpRequest,
        queryset: QuerySet[fc_models.Form],
    ) -> HttpResponse:
        """Export questions to a CSV file."""
        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = "attachment; filename=questions.csv"
        fc_exporters.export_questions(
            fc_models.FormQuestion.objects.filter(
                form_id__in=queryset.values_list("id", flat=True)
            ),
            response,
        )
        return response

    @admin.action(description="Export responses")
    def export_responses(
        self,
        request: HttpRequest,
        queryset: QuerySet[fc_models.Form],
    ) -> HttpResponse:
        """Export responses to a CSV file."""
        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = "attachment; filename=responses.csv"
        fc_exporters.export_responses(
            fc_models.FormResponse.objects.filter(
                form_responder__form_id__in=queryset.values_list(
                    "id", flat=True
                )
            ),
            response,
        )
        return response


@admin.register(fc_models.FormResponder)
class FormResponderAdmin(admin.ModelAdmin):
    list_display = ("form", "user", "created_dt")
    search_fields = (
        "form__title",
        "user__username",
        "user__email",
        "user__first_name",
        "user__last_name",
    )
    date_hierarchy = "created_dt"
    raw_id_fields = ("form", "user")
    inlines = (FormResponseInline,)



from django import template
from django.contrib.auth import get_user_model
from .. import models as fc_models

User = get_user_model()
register = template.Library()


@register.filter
def can_edit_form(form: fc_models.Form, user: User) -> bool:
    """Check if the user can edit the form."""
    return form.can_edit(user)


@register.filter
def can_delete_form(form: fc_models.Form, user: User) -> bool:
    """Check if the user can delete the form."""
    return form.can_delete(user)


@register.filter
def can_complete_form(form: fc_models.Form, user: User) -> bool:
    """Check if the user can complete the form."""
    return form.can_complete_form(user)


@register.simple_tag
def form_completed_on(form: fc_models.Form, user: User) -> str:
    """Get the date the user completed the form."""
    responder = form.completed_by(user)
    if responder:
        return responder.created_dt.strftime("%Y-%m-%d")
    return ""



from functools import wraps
from django.shortcuts import get_object_or_404, redirect
from django.http import HttpRequest, HttpResponse
from django.contrib import messages
from django.core.exceptions import PermissionDenied
from . import models as fc_models


def with_form(can_edit=False, can_delete=False):
    """Using the `pk` and `slug` parameters, retrieve the form.
    If the user is not allowed to see the form, raise a PermissionDenied.
    If the form does not exist, raise a 404.

    :param can_edit: If True, the user must be allowed to edit the form.
    :type can_edit: bool
    :param can_delete: If True, the user must be allowed to delete the form.
    :type can_delete: bool
    """

    def decorator(func):
        @wraps(func)
        def wrapper(
            request: HttpRequest,
            pk: int,
            slug: str,
            *args,
            **kwargs,
        ) -> HttpResponse:
            """Replaces the `pk` and `slug` parameters with the form object.

            :param request: The request object.
            :param type: HttpRequest
            :param pk: The primary key of the form.
            :param type: int
            :param slug: The slug of the form.
            :param type: str
            :return: The original function with the form object as the
                parameter replacing the `pk` and `slug` parameters.
            :rtype: HttpResponse
            """

            form = get_object_or_404(fc_models.Form, pk=pk, slug=slug)
            if can_edit and not form.can_edit(request.user):
                raise PermissionDenied
            if can_delete and not form.can_delete(request.user):
                raise PermissionDenied

            return func(request, form, *args, **kwargs)

        return wrapper

    return decorator


def redirect_if_form_completed(redirect_url: str = "/"):
    """If the form has been completed, redirect to the redirect_url.

    :param redirect_url: The URL to redirect to if the form is already
        completed.
    :type redirect_url: str
    """

    def decorator(func):
        @wraps(func)
        def wrapper(
            request: HttpRequest, form: fc_models.Form, *args, **kwargs
        ):
            """If the form has been completed, redirect to the redirect_url.

            :param request: The request object.
            :param type: HttpRequest
            :param form: The form object.
            :param type: fc_models.Form
            :return: The original function if the user has not completed the
                form. Otherwise, redirect to the redirect_url.
            :rtype: HttpResponse
            """

            if form.completed_by(request.user):
                messages.error(
                    request,
                    "You have already completed this form.",
                )
                return redirect(redirect_url)

            return func(request, form, *args, **kwargs)

        return wrapper

    return decorator



from django import forms
from django.utils import timezone
from django.db import transaction
from django.db.models import QuerySet
from django.contrib.auth import get_user_model

from . import models as fc_models
from .question_form_fields import field_type_map, is_choice_field

User = get_user_model()


class NewForm(forms.ModelForm):
    """Form for creating a new form."""

    class Meta:
        model = fc_models.Form
        fields = [
            "title",
            "description",
            "start_dt",
            "end_dt",
            "owner",
            "editors",
        ]

    def __init__(self, *args, **kwargs):
        user = kwargs.pop("user")
        editor_choices = kwargs.pop("editor_choices", None)
        super().__init__(*args, **kwargs)

        self._setup_datetime_fields()
        self._set_owner_field(user)
        self._set_editors_field(editor_choices)

    def _setup_datetime_fields(self) -> None:
        """Set the `start_dt` and `end_dt` to datetime fields and set the
        `start_dt` to the current time.
        """
        start_dt = self.fields["start_dt"]
        end_dt = self.fields["end_dt"]
        start_dt.initial = timezone.now()
        start_dt.widget = forms.DateTimeInput(attrs={"type": "datetime-local"})
        end_dt.widget = forms.DateTimeInput(attrs={"type": "datetime-local"})

    def _set_owner_field(self, user: User) -> None:
        """Set the owner of the form."""
        self.fields["owner"].initial = user

    def _set_editors_field(self, editor_choices: QuerySet[User]) -> None:
        """Set the editor choices for the form if provided, otherwise hide the
        field.
        """
        editors_field = self.fields["editors"]
        editors_field.required = False
        if editor_choices:
            editors_field.queryset = editor_choices
        else:
            editors_field.queryset = User.objects.none()
            editors_field.widget = forms.HiddenInput()


class DeleteForm(forms.ModelForm):
    """Form for deleting a form."""

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop("user")
        super().__init__(*args, **kwargs)

    class Meta:
        model = fc_models.Form
        fields = ["id"]

    def clean(self) -> None:
        """Check if the user can delete the form."""
        if not self.instance.can_delete(self.user):
            raise forms.ValidationError("You cannot delete this form.")


class FormQuestionForm(forms.ModelForm):
    """Form for creating a new form question."""

    choices = forms.CharField(
        widget=forms.Textarea(attrs={"rows": 1}),
        required=False,
    )
    description = forms.CharField(
        widget=forms.Textarea(attrs={"rows": 5}),
        required=False,
    )

    class Meta:
        model = fc_models.FormQuestion
        exclude = ["form"]

    def __init__(self, form_id: int, *args, **kwargs):
        self.form_id = form_id
        super().__init__(*args, **kwargs)

    def save(self, *args, **kwargs) -> fc_models.FormQuestion:
        """Save the form question and set the form id."""
        kwargs.pop("commit", None)
        form_question = super().save(commit=False, *args, **kwargs)
        form_question.form_id = self.form_id
        form_question.save()
        return form_question


class CaptureResponseForm(forms.Form):
    """Form for capturing a form response."""

    def __init__(self, form: fc_models.Form, *args, **kwargs):
        self.form = form
        super().__init__(*args, **kwargs)
        self._setup_fields()

    def _setup_fields(self) -> None:
        """Set up the fields for the form response."""
        for question in self.form.questions.all():
            self._add_field(question)

    def _add_field(self, question: fc_models.FormQuestion) -> None:
        """Add a field for the question."""
        field_type = field_type_map[question.field_type]
        field_name = f"question_{question.id}"
        field_kwargs = {
            "label": question.question,
            "required": question.required,
            "help_text": question.description,
        }
        if is_choice_field(field_type):
            choices = question.choices.split("|")
            field_kwargs["choices"] = [(c, c) for c in choices]
        self.fields[field_name] = field_type(**field_kwargs)

    def save(self, user: User, *args, **kwargs) -> fc_models.FormResponder:
        """Save the form response."""
        with transaction.atomic():
            form_responder = fc_models.FormResponder.objects.create(
                form=self.form,
                user=user,
            )
            form_responder.save()
            for question, answer in self.cleaned_data.items():
                fc_models.FormResponse.objects.create(
                    form_responder=form_responder,
                    question_id=question.lstrip("question_"),
                    answer=answer,
                )

        return form_responder



"""This module contains methods to export data from the database."""

import csv
from django.db.models import QuerySet
from . import models as fc_models


def export_questions(form_questions: QuerySet[fc_models.FormQuestion], output):
    """Export the questions in a form to a CSV file."""
    writer = csv.writer(output)
    writer.writerow(
        [
            "Form",
            "Question",
            "Type",
            "Required",
            "Seq. No.",
            "Choices",
            "Related Question",
        ]
    )
    for question in form_questions:
        writer.writerow(
            [
                question.form.title,
                question.question,
                question.field_type,
                question.required and "Yes" or "No",
                question.seq_no,
                question.choices,
                question.related_question,
            ]
        )


def export_responses(form_responses: QuerySet[fc_models.FormResponse], output):
    """Export the responses in a form to a CSV file."""
    writer = csv.writer(output)
    writer.writerow(
        [
            "Form",
            "Username",
            "Email",
            "Answered On",
            "Question",
            "Answer",
        ]
    )
    for response in form_responses:
        writer.writerow(
            [
                response.form_responder.form,
                response.form_responder.user.username,
                response.form_responder.user.email,
                response.form_responder.created_dt,
                response.question.question,
                response.answer,
            ]
        )



from django.urls import path
from . import views

app_name = "form_creator"

urlpatterns = [
    path("forms/", views.FormListView.as_view(), name="form_list"),
    path("forms/create/", views.FormCreateView.as_view(), name="form_create"),
    path(
        "forms/<int:pk>-<slug:slug>/",
        views.FormDetailView.as_view(),
        name="form_detail",
    ),
    path(
        "forms/<int:pk>-<slug:slug>/edit/",
        views.FormUpdateView.as_view(),
        name="form_edit",
    ),
    path(
        "forms/<int:pk>-<slug:slug>/delete/",
        views.FormDeleteView.as_view(),
        name="form_delete",
    ),
    path(
        "forms/<int:pk>-<slug:slug>/response/",
        views.FormResponseView.as_view(),
        name="form_response",
    ),
    path(
        "forms/<int:pk>-<slug:slug>/export/questions/",
        views.download_questions,
        name="download_questions",
    ),
    path(
        "forms/<int:pk>-<slug:slug>/export/responses/",
        views.download_responses,
        name="download_responses",
    ),
    path(
        "forms/<int:pk>-<slug:slug>/questions/edit/",
        views.FormQuestionsEditView.as_view(),
        name="form_questions_edit",
    ),
]



from types import SimpleNamespace
from django.test import TestCase, Client
from django.urls import reverse
from django.http import Http404
from django.contrib.messages import get_messages
from django.contrib.auth import get_user_model
import mock
from model_bakery import baker
from .. import views as fc_views, models as fc_models
from ..question_form_fields import FieldTypeChoices

User = get_user_model()


class TestFormSingleItemMixin(TestCase):
    def test_get_object(self):
        """Test that the `get_object` method returns the correct object."""
        form_instance = baker.make(fc_models.Form)

        self_obj = SimpleNamespace(
            request=SimpleNamespace(
                path=f"/forms/{form_instance.id}-{form_instance.slug}/",
            ),
            model=fc_models.Form,
        )
        self.assertEqual(
            fc_views.FormSingleItemMixin.get_object(self_obj),
            form_instance,
        )

    def test_get_object_invalid(self):
        self_obj = SimpleNamespace(
            request=SimpleNamespace(
                path="/forms/1-abc/",
            ),
            model=fc_models.Form,
        )
        with self.assertRaises(Http404):
            fc_views.FormSingleItemMixin.get_object(self_obj)


class TestFormCreateView(TestCase):
    def setUp(self):
        self.client = Client()
        self.user = baker.make(User)
        self.client.force_login(self.user)

    def test_form_create_view_loads(self):
        """Test that the form actually loads."""
        response = self.client.get(reverse("form_creator:form_create"))
        self.assertEqual(response.status_code, 200)

    def test_get_form_kwargs(self):
        """Test the `get_form_kwargs` method loads the editors when
        provided.
        """
        baker.make(User, _quantity=2)
        editors = User.objects.all()

        with mock.patch.object(
            fc_views.FormBaseView, "editor_choices", editors
        ):
            response = self.client.get(reverse("form_creator:form_create"))
            self.assertEqual(response.status_code, 200)
            self.assertEqual(
                set(
                    response.context["form"]
                    .fields["editors"]
                    .queryset.values_list("id")
                ),
                set(editors.values_list("id")),
            )


class TestFormListView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.owner = baker.make(User)
        cls.editor = baker.make(User)
        cls.owners_form = baker.make(fc_models.Form, owner=cls.owner)
        cls.editors_form = baker.make(fc_models.Form, editors=[cls.editor])
        baker.make(fc_models.Form)

    @staticmethod
    def self_obj(user: User):
        return SimpleNamespace(
            request=SimpleNamespace(
                user=user,
            ),
            model=fc_models.Form,
        )

    def test_get_queryset_staff(self):
        """Test that the `get_queryset` method returns all forms for staff."""
        self_obj = self.self_obj(baker.make(User, is_staff=True))
        self.assertEqual(
            fc_views.FormListView.get_queryset(self_obj).count(),
            3,
        )

    def test_get_queryset_owner(self):
        """Test that the `get_queryset` method returns only the owner's
        forms.
        """
        self_obj = self.self_obj(self.owner)
        result = fc_views.FormListView.get_queryset(self_obj)
        self.assertEqual(result.count(), 1)
        self.assertEqual(result.first(), self.owners_form)

    def test_get_queryset_editor(self):
        """Test that the `get_queryset` method returns only the editor's
        forms.
        """
        self_obj = self.self_obj(self.editor)
        result = fc_views.FormListView.get_queryset(self_obj)
        self.assertEqual(result.count(), 1)
        self.assertEqual(result.first(), self.editors_form)


class TestFormDeleteView(TestCase):
    def setUp(self):
        self.client = Client()
        self.user = baker.make(User)
        self.client.force_login(self.user)

    def test_get_view_loads(self):
        """Test that the form actually loads."""
        form_instance = baker.make(fc_models.Form, owner=self.user)
        response = self.client.get(
            reverse(
                "form_creator:form_delete",
                kwargs={
                    "pk": form_instance.id,
                    "slug": form_instance.slug,
                },
            )
        )
        self.assertEqual(response.status_code, 200)

    def test_get_view_403(self):
        """Test that the form returns a 403 if the user is not the owner."""
        form_instance = baker.make(fc_models.Form)
        response = self.client.get(
            reverse(
                "form_creator:form_delete",
                kwargs={
                    "pk": form_instance.id,
                    "slug": form_instance.slug,
                },
            )
        )
        self.assertEqual(response.status_code, 403)


class TestFromQuestionEditView(TestCase):
    """Tests the `FormQuestionEditView` class."""

    @classmethod
    def setUpTestData(cls):
        cls.user = baker.make(User)
        cls.form = baker.make(fc_models.Form, owner=cls.user)

    def setUp(self):
        self.client = Client()
        self.client.force_login(self.user)
        self.question_1 = baker.make(
            fc_models.FormQuestion,
            form=self.form,
        )
        self.question_2 = baker.make(
            fc_models.FormQuestion,
            form=self.form,
        )

    def test_get_view_loads(self):
        """Test that the form actually loads."""
        response = self.client.get(
            reverse(
                "form_creator:form_questions_edit",
                kwargs={
                    "pk": self.question_1.form.id,
                    "slug": self.question_1.form.slug,
                },
            )
        )
        self.assertEqual(response.status_code, 200)

    def test_post_view(self):
        """Test a post request with one change and one delete."""
        self.client.post(
            reverse(
                "form_creator:form_questions_edit",
                kwargs={
                    "pk": self.question_1.form.id,
                    "slug": self.question_1.form.slug,
                },
            ),
            data={
                "form-TOTAL_FORMS": "2",
                "form-INITIAL_FORMS": "2",
                "form-MIN_NUM_FORMS": "0",
                "form-MAX_NUM_FORMS": "1000",
                "form-0-id": str(self.question_1.id),
                "form-0-question": "question 1",
                "form-0-field_type": "text",
                "form-0-seq_no": "1",
                "form-1-DELETE": "on",
                "form-1-id": str(self.question_2.id),
            },
            instance=self.question_1,
        )

        self.assertEqual(fc_models.FormQuestion.objects.count(), 1)
        self.assertFalse(
            fc_models.FormQuestion.objects.filter(
                id=self.question_2.id,
            ).exists()
        )

    def test_post_view_not_changed(self):
        """Test a post request with no changes."""
        res = self.client.post(
            reverse(
                "form_creator:form_questions_edit",
                kwargs={
                    "pk": self.question_1.form.id,
                    "slug": self.question_1.form.slug,
                },
            ),
            data={
                "form-TOTAL_FORMS": "1",
                "form-INITIAL_FORMS": "1",
                "form-MIN_NUM_FORMS": "0",
                "form-MAX_NUM_FORMS": "1000",
                "form-0-id": str(self.question_1.id),
                "form-0-question": self.question_1.question,
                "form-0-field_type": self.question_1.field_type,
                "form-0-seq_no": self.question_1.seq_no,
            },
        )

        self.assertEqual(
            res.url,
            reverse(
                "form_creator:form_detail",
                kwargs={"pk": self.form.id, "slug": self.form.slug},
            ),
        )

    def test_post_view_with_errors(self):
        """Test a post request with errors."""
        res = self.client.post(
            reverse(
                "form_creator:form_questions_edit",
                kwargs={
                    "pk": self.question_1.form.id,
                    "slug": self.question_1.form.slug,
                },
            ),
            data={
                "form-TOTAL_FORMS": "1",
                "form-INITIAL_FORMS": "1",
                "form-MIN_NUM_FORMS": "0",
                "form-MAX_NUM_FORMS": "1000",
                "form-0-id": str(self.question_1.id),
                "form-0-question": "",
                "form-0-field_type": "zzzz",
                "form-0-seq_no": "1",
            },
        )

        self.assertEqual(res.status_code, 200)
        messages = list(get_messages(res.wsgi_request))
        self.assertEqual(len(messages), 1)
        self.assertEqual(
            str(messages[0]),
            "Please correct the errors below.",
        )


class TestFormResponseView(TestCase):
    """Tests the `FormResponseView` class."""

    @classmethod
    def setUpTestData(cls):
        cls.user = baker.make(User)
        cls.form = baker.make(fc_models.Form)
        cls.text_q = baker.make(
            fc_models.FormQuestion,
            form=cls.form,
            required=True,
        )
        cls.choice_q = baker.make(
            fc_models.FormQuestion,
            form=cls.form,
            field_type=FieldTypeChoices.CHOICE,
            choices="a|b|c",
        )

    def setUp(self):
        self.client = Client()
        self.client.force_login(self.user)

    def view_url(self) -> str:
        """Return the URL for the view."""
        return reverse(
            "form_creator:form_response",
            kwargs={
                "pk": self.form.id,
                "slug": self.form.slug,
            },
        )

    def test_get_view_loads(self):
        """Test that the form actually loads."""
        response = self.client.get(self.view_url())
        self.assertEqual(response.status_code, 200)

    def test_redirect_completed(self):
        """If the form is completed, redirect the user."""
        baker.make(fc_models.FormResponder, form=self.form, user=self.user)
        response = self.client.get(self.view_url())
        self.assertEqual(response.status_code, 302)

    def test_post(self):
        """Test a post request."""
        response = self.client.post(
            self.view_url(),
            data={
                f"question_{self.text_q.id}": "text answer",
                f"question_{self.choice_q.id}": "b",
            },
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(fc_models.FormResponder.objects.count(), 1)

        form_response = fc_models.FormResponder.objects.first()

        self.assertEqual(form_response.responses.count(), 2)
        self.assertEqual(
            set(
                form_response.responses.all().values_list(
                    "question_id", flat=True
                )
            ),
            {self.text_q.id, self.choice_q.id},
        )
        self.assertEqual(
            form_response.responses.get(question=self.text_q).answer,
            "text answer",
        )
        self.assertEqual(
            form_response.responses.get(question=self.choice_q).answer,
            "b",
        )

    def test_post_invalid_form(self):
        """Test a post request with invalid form."""
        response = self.client.post(
            self.view_url(),
            data={
                f"question_{self.text_q.id}": "",
                f"question_{self.choice_q.id}": "",
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(fc_models.FormResponder.objects.count(), 0)


class TestDownloadQuestions(TestCase):
    """Tests the `download_questions view."""

    @classmethod
    def setUpTestData(cls):
        cls.user = baker.make(User)
        cls.form = baker.make(fc_models.Form, owner=cls.user)
        cls.text_q = baker.make(
            fc_models.FormQuestion,
            form=cls.form,
            required=True,
        )
        cls.choice_q = baker.make(
            fc_models.FormQuestion,
            form=cls.form,
            field_type=FieldTypeChoices.CHOICE,
            choices="a|b|c",
        )

    def setUp(self):
        self.client = Client()
        self.client.force_login(self.user)

    def test_loads(self):
        """Test that the form actually loads."""
        response = self.client.get(
            reverse(
                "form_creator:download_questions",
                kwargs={
                    "pk": self.form.id,
                    "slug": self.form.slug,
                },
            )
        )
        self.assertEqual(response.status_code, 200)


class TestDownloadResponses(TestCase):
    """Tests the `download_responses view."""

    @classmethod
    def setUpTestData(cls):
        cls.user = baker.make(User)
        cls.form = baker.make(fc_models.Form, owner=cls.user)
        cls.text_q = baker.make(
            fc_models.FormQuestion,
            form=cls.form,
            required=True,
        )
        cls.choice_q = baker.make(
            fc_models.FormQuestion,
            form=cls.form,
            field_type=FieldTypeChoices.CHOICE,
            choices="a|b|c",
        )

    def setUp(self):
        self.client = Client()
        self.client.force_login(self.user)

    def test_loads(self):
        """Test that the form actually loads."""
        response = self.client.get(
            reverse(
                "form_creator:download_responses",
                kwargs={
                    "pk": self.form.id,
                    "slug": self.form.slug,
                },
            )
        )
        self.assertEqual(response.status_code, 200)



from django.core.exceptions import PermissionDenied
from django.http import HttpResponse, Http404
from django.test import TestCase
from django.test.client import RequestFactory
from model_bakery import baker
from .. import decorators, models as fc_models


class TestWithFormDecorator(TestCase):
    """Tests for the `with_form` decorator."""

    def setUp(self):
        self.user = baker.make(fc_models.User)
        self.factory = RequestFactory()
        self.form = baker.make(fc_models.Form)

    def test_object_does_not_exist(self):
        """Test that a 404 is raised when the form does not exist."""

        @decorators.with_form()
        def a_view(request, pk, slug):
            return HttpResponse()

        request = self.factory.get("/")
        request.user = self.user
        with self.assertRaises(Http404):
            a_view(request, pk=1, slug="abc")

    def test_object_does_exist(self):
        """Test that the form is returned when the form exists."""

        @decorators.with_form()
        def a_view(request, form):
            return HttpResponse()

        request = self.factory.get("/")
        request.user = self.user
        response = a_view(request, pk=self.form.pk, slug=self.form.slug)
        self.assertEqual(response.status_code, 200)

    def test_can_can_passes(self):
        """Test that the view is loaded for a user who has editable permissions
        when this is required.
        """

        @decorators.with_form(can_edit=True)
        def a_view(request, form):
            return HttpResponse()

        request = self.factory.get("/")
        request.user = self.user
        self.form.owner = self.user
        self.form.save()

        response = a_view(request, pk=self.form.pk, slug=self.form.slug)
        self.assertEqual(response.status_code, 200)

    def test_can_edit_fails(self):
        """Test that the view is not loaded for a user who does not have
        editable permissions when this is required.
        """

        @decorators.with_form(can_edit=True)
        def a_view(request, form):
            return HttpResponse()

        request = self.factory.get("/")
        request.user = self.user

        with self.assertRaises(PermissionDenied):
            a_view(request, pk=self.form.pk, slug=self.form.slug)

    def test_can_delete_passes(self):
        """Test that the view is loaded for a user who has delete permissions
        when this is required.
        """

        @decorators.with_form(can_delete=True)
        def a_view(request, form):
            return HttpResponse()

        request = self.factory.get("/")
        request.user = self.user
        self.form.owner = self.user
        self.form.save()

        response = a_view(request, pk=self.form.pk, slug=self.form.slug)
        self.assertEqual(response.status_code, 200)

    def test_can_delete_fails(self):
        """Test that the view is not loaded for a user who does not have
        delete permissions when this is required.
        """

        @decorators.with_form(can_delete=True)
        def a_view(request, form):
            return HttpResponse()

        request = self.factory.get("/")
        request.user = self.user

        with self.assertRaises(PermissionDenied):
            a_view(request, pk=self.form.pk, slug=self.form.slug)



"""This module contains tests for the `exporters` module."""

from django.test import TestCase
from django.http import HttpResponse
from model_bakery import baker
from .. import models as fc_models, exporters as fc_exporters


class TestExportQuestions(TestCase):
    """Tests for the `export_questions` function."""

    def setUp(self):
        baker.make(fc_models.FormQuestion)
        baker.make(fc_models.FormQuestion)
        self.response = HttpResponse(content="text/csv")
        self.response["Content-Disposition"] = "attachment; filename=test.csv"

    def test_export_questions(self):
        """Test that questions are exported to a CSV file."""
        fc_exporters.export_questions(
            fc_models.FormQuestion.objects.all(), self.response
        )

        content = self.response.content.decode("utf-8")
        for form_question in fc_models.FormQuestion.objects.all():
            self.assertIn(form_question.form.title, content)
            self.assertIn(form_question.question, content)
            self.assertIn(form_question.field_type, content)
            self.assertIn(str(form_question.seq_no), content)


class TestExportResponses(TestCase):
    """Tests for the `export_responses` function."""

    def setUp(self):
        baker.make(fc_models.FormResponse)
        baker.make(fc_models.FormResponse)
        self.response = HttpResponse(content="text/csv")
        self.response["Content-Disposition"] = "attachment; filename=test.csv"

    def test_export_responses(self):
        """Test that responses are exported to a CSV file."""
        fc_exporters.export_responses(
            fc_models.FormResponse.objects.all(), self.response
        )

        content = self.response.content.decode("utf-8")
        for form_response in fc_models.FormResponse.objects.all():
            self.assertIn(form_response.form_responder.form.title, content)
            self.assertIn(form_response.form_responder.user.username, content)
            self.assertIn(
                str(form_response.form_responder.created_dt), content
            )
            self.assertIn(form_response.question.question, content)



"""Tests for the `admin` module."""

from django.test import TestCase, Client
from django.urls import reverse
from django.http import HttpResponse
from model_bakery import baker
from .. import admin as fc_admin, models as fc_models


class TestFormQuestionInline(TestCase):
    """Tests for the `FormQuestionInline` class."""

    @classmethod
    def setUpTestData(cls):
        cls.user = baker.make(fc_models.User, is_superuser=True, is_staff=True)
        cls.form = baker.make(fc_models.Form)
        cls.questions = baker.make(
            fc_models.FormQuestion,
            form=cls.form,
            _quantity=2,
        )
        baker.make(fc_models.FormQuestion)

    def setUp(self):
        self.client = Client()
        self.client.force_login(user=self.user)

    def change_form_url(self) -> str:
        """Return the URL to the admin page to edit the form."""
        return reverse("admin:form_creator_form_change", args=[self.form.id])

    @staticmethod
    def add_form_url() -> str:
        """Return the URL to the admin page to add a form."""
        return reverse("admin:form_creator_form_add")

    def test_related_questions_change_form(self):
        """Test that the related questions queryset presents questions related
        to this form.
        """
        response = self.client.get(self.change_form_url())

        self.assertEqual(
            (
                response.context["inline_admin_form"]
                .formset.forms[0]
                .fields["related_question"]
                .queryset.count()
            ),
            len(self.questions),
        )

    def test_related_questions_add_form(self):
        """Test that the related questions queryset provides no options when
        adding a new form.
        """

        self.assertEqual(
            (
                self.client.get(self.add_form_url())
                .context[0]["inline_admin_formsets"][0]
                .formset.form()
                .fields["related_question"]
                .queryset.count()
            ),
            0,
        )


class TestAdminForm(TestCase):
    """Tests for the `FormAdmin` class."""

    @classmethod
    def setUpTestData(cls):
        cls.form = baker.make(fc_models.Form)
        cls.question = baker.make(
            fc_models.FormQuestion,
            form=cls.form,
            _quantity=2,
        )
        cls.responses = baker.make(
            fc_models.FormResponse,
            form_responder__form=cls.form,
            _quantity=2,
        )

    def test_export_questions(self):
        """Test that the export questions view returns a CSV file."""
        questions = fc_admin.FormAdmin.export_questions(
            None, None, fc_models.Form.objects.filter(id=self.form.id)
        )
        self.assertIsInstance(questions, HttpResponse)
        self.assertEqual(questions["Content-Type"], "text/csv")

    def test_export_responses(self):
        """Test that the export responses view returns a CSV file."""
        responses = fc_admin.FormAdmin.export_responses(
            None, None, fc_models.Form.objects.filter(id=self.form.id)
        )
        self.assertIsInstance(responses, HttpResponse)
        self.assertEqual(responses["Content-Type"], "text/csv")
